/*
 * Interrupt and TLB miss handling support.
 *
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

.text
.extern interrupt_handler
.globl _tlb_entry
.globl _irq_entry
.globl _end_entries
.set noreorder

/* NOTE: Duplicated from cpu.h. */

#define page_table_start 0x00040000
#define page_table_task_size 0x00008000
#define page_table_task_size_log2 15

_tlb_entry:
	/* Get the bad address. */

	mfc0 $k0, $10				/* CP0_ENTRYHI */
	nop
	andi $k1, $k0, 0xff			/* ASID */

	/* For ASID == 0... */

	beqz $k1, _tlb_entry_direct
	nop

	/* For addresses over 0x00080000... */

	li $k1, 0xfff80000
	and $k1, $k0, $k1
	bnez $k1, _tlb_entry_direct
	nop

	/* Otherwise, load the page table entries. */

	andi $k1, $k0, 0xff			/* ASID */
	sll $k1, $k1, page_table_task_size_log2	/* [ASID] */
	li $k0, page_table_start		/* page_table */
	add $k0, $k0, $k1			/* page_table[ASID] */

	mfc0 $k1, $4				/* CP0_CONTEXT */
	nop
	srl $k1, $k1, 1				/* use 8 byte - not 16 byte - entries */
	add $k0, $k0, $k1			/* page_table[ASID][entry] */

	lw $k1, 0($k0)				/* page_table[ASID][entry][0] */
	mtc0 $k1, $2				/* CP0_ENTRYLO0 */
	lw $k1, 4($k0)				/* page_table[ASID][entry][1] */
	mtc0 $k1, $3				/* CP0_ENTRYLO1 */
						/* page size is 4KB */
	mtc0 $zero, $5				/* CP0_PAGEMASK */
	nop

	tlbwr
	nop
	eret
	nop

_tlb_entry_direct:
	/* Otherwise, just translate the address directly. */

	li $k1, 0xffffe000
	and $k0, $k0, $k1			/* VPN2 (8KB resolution) */
	srl $k0, $k0, 6				/* PFN (maintain 8KB resolution, bit 6 remaining zero) */
	ori $k0, $k0, 0x1e			/* flags */

	mtc0 $k0, $2				/* CP0_ENTRYLO0 */
	ori $k0, $k0, 0x40			/* page size is 4KB (bit 6 set) */
	mtc0 $k0, $3				/* CP0_ENTRYLO1 */
	nop					/* page size is 4KB */
	mtc0 $zero, $5				/* CP0_PAGEMASK */
	nop

	tlbwr
	nop
	eret
	nop

_irq_entry:
	/* Save the status. */

	mfc0 $k0, $12				/* CP0_STATUS */
	nop
	sw $k0, -120($sp)

	/* Mask interrupts. */

	li $k1, 0xffff03ff
	and $k1, $k0, $k1
	mtc0 $k1, $12				/* CP0_STATUS */

	/* Save registers that the assembler wants to trash. */

        sw $t9, -100($sp)
        sw $gp, -104($sp)
        sw $ra, -112($sp)

	lui $gp, %hi(_GLOBAL_OFFSET_TABLE_)
	ori $gp, $gp, %lo(_GLOBAL_OFFSET_TABLE_)
	la $k0, interrupt_handler
	jr $k0
	nop

_end_entries:

.set reorder
